"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.elink.elink_parameter
  :synopsis:
    Exports class ElinkParameters  for NCBI E-Utils queries.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import sys
import json


import entrezpy.base.parameter
import entrezpy.log.logger


class ElinkParameter(entrezpy.base.parameter.EutilsParameter):
  """ElinkParameter checks query specific parameters and configures a
  :class:`entrezpy.elink.elink_query.ElinkQuery` instance. A link gets its size
  from :attr:`entrezpy.elink.elink_parameter.ElinkParameter.uids` (from the `id`
  Eutils parameter) or earlier result stored on the Entrez history server.
  :attr:`entrezpy.elink.elink_parameter.ElinkParameter.retmode` is JSON where
  possible and :attr:`entrezpy.elink.elink_parameter.ElinkParameter.cmd` is
  `neighbor`. ELink has no set maximum for UIDs which can be linked in one
  query, fixing :attr:`entrezpy.elink.elink_parameter.ElinkParameter.query_size`,
  :attr:`entrezpy.elink.elink_parameter.ElinkParameter.request_size`, and
  :attr:`entrezpy.elink.elink_parameter.ElinkParameter.expected_requests` to 1.

  :param dict parameter: Eutils Elink parameter
  """

  nodb_cmds = {'acheck', 'ncheck', 'lcheck', 'llinks', 'llinkslib', 'prlinks'}
  """Elink commands not requiring the `db` parameter"""

  retmodes = {'llinkslib' : 'xml'}
  """The llinkslib elink command is the only command only returning XML"""

  def_retmode = 'json'
  """Use JSON whenever possible"""


  def __init__(self, parameter):
    super().__init__(parameter)
    self.cmd = parameter.get('cmd', 'neighbor')
    self.dbfrom = parameter.get('dbfrom')
    self.uids = parameter.get('id', [])
    self.retmode = self.set_retmode(parameter.get('retmode'))
    self.linkname = parameter.get('linkname')
    self.term = parameter.get('term')
    self.holding = parameter.get('holding')
    self.datetype = parameter.get('datetype')
    self.reldate = parameter.get('reldate')
    self.mindate = parameter.get('mindate')
    self.maxdate = parameter.get('maxdate')
    self.doseq = parameter.get('link', True)
    self.query_size = 1
    self.request_size = 1
    self.expected_requests = 1
    self.check()
    self.logger = entrezpy.log.logger.get_class_logger(ElinkParameter)
    self.logger.debug(json.dumps({'init':self.dump()}))

  def check(self):
    """
    Implements :meth:`entrezpy.base.parameter.check` and aborts if required
    parameters  are missing.
    """
    if self.cmd not in ElinkParameter.nodb_cmds and not self.haveDb():
      sys.exit(self.logger.error(json.dumps({'Missing required parameters':'db/cmd',
                                             'action':'abort'})))
    if not self.dbfrom:
      sys.exit(self.logger.error(json.dumps({'Missing required parameter':'dbfrom',
                                             'action':'abort'})))

    if not self.uids and not self.haveWebenv and not self.haveQuerykey:
      sys.exit(self.logger.error(json.dumps({'Missing required parameters': {'ids':self.uids,
                                                                             'QueryKey':self.querykey,
                                                                             'WebEnv':self.webenv},
                                             'action':'abort'})))

  def set_retmode(self, retmode):
    """Checks for valid and supported Elink retmodes

    :param str retmode: requested retmode
    :return: default or cmd adjusted cretmode
    :rtype: str
    """
    if retmode == 'ref':
      self.logger.warning(json.dumps({'retmode ref':'not supported',
                                        'using':ElinkParameter.def_retmode}))
      return ElinkParameter.def_retmode
    return ElinkParameter.retmodes.get(self.cmd, ElinkParameter.def_retmode)

  def dump(self):
    """:return: Instance attributes
    :rtype: dict
    """
    return {'db' : self.db,
            'WebEnv':self.webenv,
            'query_key' : self.querykey,
            'dbfrom' : self.dbfrom,
            'cmd' : self.cmd,
            'uids' : self.uids,
            'retmode' : self.retmode,
            'linkname' : self.linkname,
            'term' : self.term,
            'holding' : self.holding,
            'datetype' : self.datetype,
            'doseq' : self.doseq,
            'reldate' : self.reldate,
            'mindate' : self.mindate,
            'maxdate' : self.maxdate,
            'query_size' : self.query_size,
            'request_size' : self.request_size,
            'expected_requets' : self.expected_requests}



"""
..
  Copyright 2020 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.log.conf
  :synopsis: Stores default logging configuration.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


default_config = {
  'disable_existing_loggers': False,
  'version': 1,
  'formatters':
  {
    'default':
    {
      'format': '%(asctime)s %(threadName)s [%(levelname)s] %(name)s: %(message)s'
    },
  },
  'handlers':
  {
    'console':
    {
      'class': 'logging.StreamHandler',
      'stream' : 'ext://sys.stderr',
      'formatter': 'default'
    }
  },
  'loggers':
  {
    '':
    {
      'handlers': ['console']
    }
  }
}
"""Dictionary to store logger configuration"""



"""
..
  Copyright 2018 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.efetch.efetch_analyzer
  :synopsis: Exports the class EfetchAnalyzer implementing the entrezpy analysis
    of Efetch Eutils results from NCBI Eutils queries

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import json


import entrezpy.base.analyzer
import entrezpy.log.logger


class EfetchAnalyzer(entrezpy.base.analyzer.EutilsAnalyzer):
  """EfetchAnalyzer implements a basic analysis of Efetch E-Utils responses.
  Stores results in a :class:`entrezpy.efetch.efetch_result.EfetchResult`
  instance.

  .. note:: This is a very superficial analyzer for documentation and
    educational purposes. In almost all cases a more specific analyzer has to be
    implemented in inheriting :class:`entrezpy.base.analyzer.EutilsAnalyzer`
    and implementing the virtual functions
    :meth:`entrezpy.base.analyzer.EutilsAnalzyer.analyze_result` and
    :meth:`entrezpy.base.analyzer.EutilsAnalzyer.analyze_error`.
  """

  def __init__(self):
    """:ivar result: :class:`entrezpy.efetch.efetch_result.EfetchResult`"""
    super().__init__()
    self.result = None
    self.logger = entrezpy.log.logger.get_class_logger(EfetchAnalyzer)

  def init_result(self, response, request):
    """Should be implemented if used properly"""
    if not self.result:
      self.result = True
      print(self.norm_response(response, request.rettype))
      return True
    return False

  def analyze_result(self, response, request):
    if not self.init_result(response, request):
      print(self.norm_response(response, request.rettype))

  def analyze_error(self, response, request):
    self.logger.error(json.dumps({'response':self.norm_response(response, request.rettype)}))

    self.logger.debug(json.dumps({'response':{'dump':request.dump_internals(),
                                              'error':self.norm_response(response, request.rettype)}}))

  def norm_response(self, response, rettype=None):
    """Normalizes response for printing

    :param response: efetch response
    :type  response: dict or `io.StringIO`
    :return: str or dict
    """
    if rettype == 'json':
      return response
    return response.getvalue()

  def isEmpty(self):
    if not self.result:
      return True
    return False



"""
..
  Copyright 2018 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.esummary.esummary_analyzer
   :synopsis: Exports class EsummaryAnalyzer implementing entrezpy Esummary
    queries to NCBI E-Utility Esummary

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import json

import entrezpy.base.analyzer
import entrezpy.esummary.esummary_result
import entrezpy.log.logger


class EsummaryAnalyzer(entrezpy.base.analyzer.EutilsAnalyzer):
  """
  EsummaryAnalyzer implements the analysis of ESsummary responses from E-Utils.
  JSON formatted data is enforced in responses. Summaries are stored in a
  :class:`entrezpy.esummary.esummary_result.EsummaryResult` instance.
  """

  def __init__(self):
    """:ivar result: Esummary results
       :type result: :class:`entrezpy.esummary.esummary_result.EsummaryResult`
    """
    super().__init__()
    self.result = None
    self.logger = entrezpy.log.logger.get_class_logger(EsummaryAnalyzer)

  def init_result(self, response, request):
    """
    Inits :attr:`.result` as
    :class:`entrezpy.esummary.esummary_result.EsummaryResult`.

    :return: if result is initiated
    :rtype: bool
    """
    if not self.result:
      self.result = entrezpy.esummary.esummary_result.EsummaryResult(response, request)
      return True
    return False

  def analyze_result(self, response, request):
    if not self.init_result(response, request):
      self.result.add_summaries(response.pop('result', None))

  def analyze_error(self, response, request):
    log_msg = {'tool':request.tool, 'request':request.id, 'query':request.query_id}
    if 'error' in response:
      log_msg.update({'error' : response.pop('error')})
    if 'esummaryresult' in response:
      log_msg.update({'error' : response.pop('esummaryresult')})
    self.logger.error(json.dumps(
      {'response':log_msg, 'dump':request.dump_internals()}))



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.base.parameter
  :synopsis: Export the base class for entrezpy parameters in NCBI E-Utils
    queries

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import sys


class EutilsParameter:
  """ EutilsParameter set and check parameters for each query. EutilsParameter
  is populated from a dictionary with valid E-Utilities parameters for the
  corresponding query. It declares virtual functions where necessary.

  Simple helper functions are presented to test the common parameters `db`,
  `WebEnv`, `query_key` and `usehistory`.

  .. note::  :attr:`.usehistory` is the parameter used for Entrez history
    queries and is set to `True` (use it) by default. It can be set to `False`
    to ommit history server use.

  :meth:`.haveExpectedRequests` tests if the  of the number of requests has been
  calculated.

  The virtual methods :meth:`.check` and :meth:`.dump` need thrir own
  implementation since they can vary between queries.

  .. warning:: :meth:`.check` is expected to run after all parameters have been
    set.

  :param dict parameter: Eutils query parameters
  """

  def __init__(self, parameter=None):
    """:ivar str db: Entrez database name
    :ivar str webenv: WebEnv
    :ivar int querykey: querykey
    :ivar int expected_request: number of expected request for the query
    :ivar bool doseq: use `id=` parameter for each uid in POST
    """
    if not parameter:
      sys.exit("Implementation Error: Require Eutils query parameters")
    self.db = parameter.get('db')
    self.webenv = parameter.get('WebEnv')
    self.querykey = parameter.get('query_key', 0)
    self.expected_requests = 1
    self.doseq = False
    self.usehistory = True

  def haveDb(self):
    """Check for required db parameter

    :rtype: bool
    """
    if self.db:
      return True
    return False

  def haveWebenv(self):
    """Check for required WebEnv parameter

    :rtype: bool
    """
    if self.webenv:
      return True
    return False

  def haveQuerykey(self):
    """Check for required QueryKey parameter

    :rtype: bool
    """
    if self.querykey:
      return True
    return False

  def useHistory(self):
    """Check if history server should be used.

    :rtype: bool
    """
    if self.usehistory:
      return True
    return False


  def haveExpectedRequets(self):
    """Check fo expected requests. Hints an error if no requests are expected.

    :rtype: bool
    """
    if self.expected_requests > 0:
      return True
    return False

  def check(self):
    """Virtual function to run a check before starting the query. This is a
    crucial step and should abort upon failing.

    :raises NotImplementedError: if not implemented
    """
    raise NotImplementedError()

  def dump(self):
    """Dump instance attributes

    :rtype: dict
    :raises NotImplementedError: if not implemented
    """
    raise NotImplementedError()



"""
..
  Copyright 2018,2019,2020 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: logger
  :synopsis: This module configures logging via Python's :mod:`logging`.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import os
import time
import logging
import logging.config


import entrezpy.log.conf


CONFIG = {'level':'INFO', 'quiet':True, 'propagate':True}
"""Store logger settings"""

def get_root():
  """Returns the module root"""
  return 'entrezpy'

def resolve_class_namespace(cls):
  """Resolves namespace for logger"""
  return f"{cls.__module__}.{cls.__qualname__}"

def get_class_logger(cls):
  """Prepares logger for given class """
  logger = logging.getLogger(f"{cls.__module__}.{cls.__qualname__}")
  logger.propagate = CONFIG['propagate']
  if CONFIG['quiet'] is True:
    logger.addHandler(logging.NullHandler())
    return logger
  logging.config.dictConfig(entrezpy.log.conf.default_config)
  logger.setLevel(CONFIG['level'])
  return logger

def set_level(level):
  """Sets logging level for applications using entrezpy."""
  CONFIG['level'] = level
  CONFIG['quiet'] = False



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.elink.linkset.unit.linklist
  :synopsis: Exports class LinkList implementing Elink results for `acheck`
    command reporting all avaible links for UIDs.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import entrezpy.elink.linkset.unit.linksetunit


class LinkList(entrezpy.elink.linkset.unit.linksetunit.LinksetUnit):
  """ The `LinkList` class represents a result from the `acheck` Elink command.
  Results show all available links for UIDs in the target database.

  :param str dbto: target database
  :param str linkname: Elink linkname
  :param str menutag: Elink menutag
  :param str htmltag: Elink htmltag
  :param int priority: Elink priority
  """

  @classmethod
  def new(cls, dbto, linkname, menutag, htmltag, priority):
    """:rtype: :class:`linksetunit.linkin.LinkList`"""
    return cls(dbto, linkname, menutag, htmltag, priority)

  def __init__(self, dbto=None, linkname=None, menutag=None, htmltag=None, priority=None):
    super().__init__(dbto, linkname)
    self.menutag = menutag
    self.htmltag = htmltag
    self.priority = priority

  def dump(self):
    """:rtype: dict"""
    return dict({'htmltag' : self.htmltag, 'priority' : self.priority}, **self.basic_dump())



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: neighbor_score
  :synopsis: Exports class NeighborScore implementing Elink results from the
    Elink `neighbor_score` command reporting similarity scores for links.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import entrezpy.elink.linkset.unit.linksetunit


class NeighborScore(entrezpy.elink.linkset.unit.linksetunit.LinksetUnit):
  """The `NeighborScore` class represents a result from the `neighbor_score`
  Elink command. Results show a goodness score for the found links.

  :param dict link: Linkscore to target database
  :param str dbto: Entrez database name for target database
  :param str linkname: Elink linkname
  """

  @classmethod
  def new(cls, link, dbto, linkname):
    """:rtype: :class:`neighbor_score.Neighbor_score`"""
    return cls(link, dbto, linkname)

  def __init__(self, link=None, dbto=None, linkname=None):
    super().__init__(dbto, linkname)
    self.uid = None if link is None else int(link['id'])
    self.score = None if link is None else  int(link['score'])

  def dump(self):
    """:rtype: dict"""
    return dict({'uid' : self.uid, 'score' : self.score}, **self.basic_dump())



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: epost_result
  :synopsis: Exports the EpostResult class implementing the results from
    Epost queries.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import json
import logging

import entrezpy.base.result
import entrezpy.log.logger


class EpostResult(entrezpy.base.result.EutilsResult):
  """ EpostResult stores WebEnv and QueryKey from posting UIDs to the
  History server. Since no limit is imposed on the number of UIDs which can be
  posted in one query, the size of the result is the size of the request and
  only one WebEnv and QueryKey are returned.

  :param request: entrezpy Epost request instance
  :request type: :class:`entrezpy.epost.epost_request.EpostRequest`
  :param dict response: response
  """

  def __init__(self, response, request):
    """:ivar list uids: posted UIDs"""
    super().__init__('epost', request.query_id, request.db, response.pop('webenv'), response.pop('querykey'))
    self.uids = request.uids
    self.logger = entrezpy.log.logger.get_class_logger(EpostResult)
    self.logger.debug({'init':self.dump()})

  def dump(self):
    return {'db':self.db, 'size':self.size(), 'len_uids':len(self.uids),
            'query_key':self.references.dump(), 'uids':self.uids,
            'function':self.function}

  def get_link_parameter(self, reqnum=0):
    return {'WebEnv':self.webenv, 'db':self.db,
            'QueryKey':self.references.get_querykey(self.webenv, reqnum)}

  def size(self):
    return len(self.uids)

  def isEmpty(self):
    if self.size() > 0:
      return False
    return True



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: neighbor_history
  :synopsis: Exports class NeighborHistory implementing Elink results from the
    Elink `neighbor_history` command repoting WebEnv and query-key for links.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import entrezpy.elink.linkset.unit.linksetunit


class NeighborHistory(entrezpy.elink.linkset.unit.linksetunit.LinksetUnit):
  """The `NeighborHistory` class represents a result from the `neighbor_history`
  Elink command. Results are returned as WebEnv and query_key.

  :param str dbto: Entrez database name for target database
  :param str linkname: Elink linkname
  :param str webenv: WebEnv for link
  :param int querykey: querykey for correspondong WebEnv
  """

  @classmethod
  def new(cls, dbto, linkname, webenv, querykey):
    """:rtype: :class:`neighbor_history.Neighbor_history`"""
    return cls(dbto, linkname, webenv, querykey)

  def __init__(self, dbto=None, linkname=None, webenv=None, querykey=None):
    super().__init__(dbto, linkname, cat='neighbor_history')
    self.webenv = webenv
    self.querykey = querykey

  def dump(self):
    """:rtype: dict"""
    return dict({'cat' : self.cat, 'webenv' : self.webenv, 'querykey' : self.querykey},
                **self.basic_dump())



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: neighbor
  :synopsis: Exports class Neighbor implementing Elink results from the Elink
    `neighbor` command reporting linked UIDs in teh target database.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import entrezpy.elink.linkset.unit.linksetunit


class Neighbor(entrezpy.elink.linkset.unit.linksetunit.LinksetUnit):
  """The `Neighbor` class represents a result from `neighbor` the Elink command.
  Neighbor results shows availbale links in teh target database.

  :param str uid: UID/accsession from the target database
  :param str dbto: Entrez database name for target database
  :param str linkname: Elink linkname
  """

  @classmethod
  def new(cls, uid, dbto, linkname):
    """:rtype: `neighbor.Neighbor`"""
    return cls(uid, dbto, linkname)

  def __init__(self, uid=None, dbto=None, linkname=None):
    super().__init__(dbto, linkname)
    self.uid = None if uid is None else int(uid)

  def dump(self):
    """:return: dict"""
    return dict({'uid' : self.uid}, **self.basic_dump())



"""
..
  Copyright 2018 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.base.monitor
  :synopsis: Exports class QueryMonitor implementing monitoring of entrezpy
    queries and requsts

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import json
import sys
import threading
import time

import entrezpy.log.logger


class QueryMonitor:
  """
  The QueryMonitor class implements the monitoring of entrezpy requests in
  entrezpy queries. It controls individual Observers which are linked to one
  entrezpy query.
  """

  logger = None
  observers = {}

  class Observer(threading.Thread):
    """
    The Observer class implements the observation of one entrezpy query. It
    uses Python's multithreading deamon when using multithreading.
    """

    def __init__(self):
      super().__init__(daemon=True)
      self.expected_requests = 0
      self.processed_requests = 0
      self.doObserve = True
      self.requests = []
      self.duration = None

    def recall(self):
      """Stops an observation for a query"""
      self.doObserve = False
      for i in self.requests:
        i.report_status(self.processed_requests, self.expected_requests)
      self.join()

    def dispatch(self, parameter):
      """
      Starts observation

      :param parameter: query parameter
      """
      self.expected_requests = parameter.expected_requests
      if not self.is_alive():
        self.start()

    def observe(self, request):
      """Adds one query request for observation"""
      self.requests.append(request)

    def run(self):
      """Observes requests from an entrezpy query"""
      while self.doObserve:
        for i in self.requests:
          i.report_status(self.processed_requests, self.expected_requests)
        time.sleep(1)

  def __init__(self, query):
    """Inits observer and registers query for monitoring."""
    self.register_query(query)
    QueryMonitor.logger = entrezpy.log.logger.get_class_logger(QueryMonitor)
    #self.locks = {}

  def register_query(self, query_id):
    """
    Adds a query for observation

    :param query: entrezpy query
    :type  query: :class:`entrezpy.base.query.EutilsQuery`
    """
    #self.locks[query_id] = threading.Lock()
    if query_id in QueryMonitor.observers:
      sys.exit(QueryMonitor.logger.error(json.dumps({'Duplicate query_id':query_id})))
    QueryMonitor.observers[query_id] = self.Observer()

  def get_observer(self, query_id):
    """
    Returns an observer for a specific query.

    :param str query_id: entrezpy query id
    :rtype: :class:`base.mnonitor.QueryMonitor.Observer`
    """
    if query_id in QueryMonitor.observers:
      return QueryMonitor.observers[query_id]
    sys.exit(QueryMonitor.logger.error(json.dumps({'No observer for query_id':query_id})))

  def dispatch_observer(self, query_id, parameter):
    """
    Start the observer for an entrezpy query

    :param query: entrezpy query
    :type  query: :class:`entrezpy.base.query.EutilsQuery`
    :param parameter: query parameter
    :type  parameter: 'class':`entrezpy.base.EutilsParameter`
    """
    QueryMonitor.logger.debug(json.dumps({'dispatching observer for':query_id,
                                          'parameter':parameter.dump()}))
    self.get_observer(query_id).dispatch(parameter)

  def recall_observer(self, query_id):
    """
    Stops the observer for an entrezpy query

    :param query: entrezpy query
    :type  query: :class:`entrezpy.base.query.EutilsQuery`
    """
    QueryMonitor.logger.debug(json.dumps({'recalling observer for':query_id}))
    self.get_observer(query_id).recall()

  def update_observer(self, query_id, parameter):
    """
    Function updating the settings for a thread. Honestly, I have no idea if
    the lock is really required. It works without locks, just updating the
    parameter.

    :param query: entrezpy query
    :type  query: :class:`entrezpy.base.query.EutilsQuery`
    :param parameter: query parameter
    :type  parameter: 'class':`entrezpy.base.EutilsParameter`
    """
    QueryMonitor.logger.debug(json.dumps({'updating observer for':query_id,
                                          'parameter':parameter.dump()}))
    self.get_observer(query_id).expected_requests = parameter.expected_requests



"""
..
  Copyright 2018 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.base.referencer
  :synopsis: Exports the EutilReferencer class managing WebEnv and query keys
    for entrezpy queries

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


class EutilReferencer:
  """EutilReferences handles History Server references for queries. WenEnv and
  corresponding query keys are stored and can be retrieved either as whole, i.e.
  Webenv and all corresponding query keys or as specific queryley for a webenv.
  Query keys for each WebeVn are stored in a list and can be retrieved as
  the N-th request for WebEnv, i.e. the first reuquest is at index 0, the last
  at -1.
  """

  class Reference:
    """Acting as struct to return all querykeys for a WebEnv"""
    def __init__(self, webenv, querykeys):
      self.webenv = webenv
      self.querykeys = querykeys

  def __init__(self, webenv, querykey):
    """:param str webenv: WebEnv
       :param int webenv: querykey
    """
    self.references = {}
    if webenv:
      self.add_reference(webenv, querykey)

  def add_reference(self, webenv, querykey):
    """Adds new refernce

    :param str webenv: WebEnv
    :param int webenv: querykey
    """
    if webenv not in self.references:
      self.references[webenv] = []
    self.references[webenv].append(int(querykey))

  def get_reference(self, webenv):
    """:param str webenv: WebEnv
       :return: History server reference for webenv
       :rtype: :class:'EutilReference.Reference' instance or None
    """
    if webenv not in self.references:
      return None
    return self.Reference(webenv, self.references[webenv])

  def get_querykey(self, webenv, querynum):
    """:return: query key for N-th query as list index, i.e. first = 0
       :rtype: int or None
    """
    if webenv not in self.references:
      return None
    if (querynum+1) > len(self.references[webenv]):
      return None
    return self.references[webenv][querynum]

  def dump(self):
    """Return all references

    :rtype: dict"""
    return self.references

  def size(self):
    """:rtype: int"""
    return len(self.references)



"""
..
  Copyright 2018 The University of Sydney

  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.base.request
  :synopsis: Exports the base class for entrezpy requests in NCBI E-Utils
    queries

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""

import json
import time

import entrezpy.log.logger


class EutilsRequest:
  """EutilsRequest is the base class for requests from
  :class:`entrezpy.base.query.EutilsQuery`.

  EutilsRequests instantiate in :meth:`entrezpy.base.query.EutilsQuery.inquire`
  before being added to the request pool by
  :meth:`entrezpy.base.query.EutilsQuery.add_request`. Each EutilsRequest
  triggers an answer at the NCBI Entrez servers if no connection errors occure.

  :class:`EutilsRequest` stores the required  information for POST requests.
  Its status can be queried from outside by
  :func:`entrezpy.base.request.EutilsRequest.get_observation`. EutilsRequest
  instances store information not present in the server response and is
  required by :class:`entrezpy.base.analyzer.EutilsAnalyzer` to parse responses
  and errors correctly. Several instance attributes are not required for a POST
  request but help debugging.

  Each request is automatoically assigned an id to identify and  trace requests
  using the query id and request id.

  :param str eutil: eutil function for this request, e.g. efetch.fcgi
  :param str db: database for request
  """

  def __init__(self, eutil, db):
    """ Initializes a new request with initial attributes as part of a query in
    :class:`entrezpy.base.query.EutilsQuery`.

    :ivar str tool: tool name to which this request belongs
    :ivar str url: full Eutil url
    :ivar str contact: use email
    :ivar str apikey: NBCI apikey
    :ivar str query_id: :attr:`entrezpy.base.query.EutilsQuery.query_id`  which
      initiated this request
    :ivar int status: request status : 0->success, 1->Fail,2->Queued
    :ivar int size: size of request, e.g. number of UIDs
    :ivar float start_time: start time of request in seconds since epoch
    :ivar duration: duration for this request in seconds
    :ivar doseq: set doseq parameter in :meth:`entrezpy.request.Request.request`

    .. note:: :attr:`.status` is work in progress.
    """
    self.eutil = eutil
    self.db = db
    self.id = None
    self.query_id = None
    self.tool = None
    self.url = None
    self.contact = None
    self.apikey = None
    self.status = 2
    self.request_error = None
    self.size = 1
    self.start_time = None
    self.duration = None
    self.doseq = True
    self.logger = entrezpy.log.logger.get_class_logger(EutilsRequest)

  def get_post_parameter(self):
    """
    Virtual function returning the POST parameters for the request from required
    attributes.

    :rtype: dict
    :raises NotImplemetedError:
    """
    raise NotImplementedError()

  def prepare_base_qry(self, extend=None):
    """
    Returns instance attributes required for every POST request.

    :param dict extend: parameters extending basic parameters
    :return: base parameters for POST request
    :rtype: dict
    """
    base = {'email' : self.contact, 'tool' : self.tool, 'db' : self.db}
    if self.apikey:
      base.update({'api_key' : self.apikey})
    if extend:
      base.update(extend)
    return base

  def set_status_success(self):
    """Set status if request succeeded"""
    self.status = 0

  def set_status_fail(self):
    """Set status if request failed"""
    self.status = 1

  def report_status(self, processed_requests=None, expected_requests=None):
    """
    Reports request status when triggered
    """
    self.logger.info(json.dumps({'query':self.query_id,
      'request':self.id, 'status':self.status}))
    self.logger.debug(json.dumps({'status': self.dump_internals()}))

  def get_request_id(self):
    """
    :returns: full request id
    :rtype: str
    """
    return '.'.join([str(self.query_id), str(self.id)])

  def set_request_error(self, error):
    """
    Sets request error and HTTP/URL error message

    :param str error: HTTP/URL error
    """
    self.request_error = error
    self.status = 1

  def start_stopwatch(self):
    """Starts time to measure request duration."""
    self.start_time = time.time()

  def calc_duration(self):
    """Calculates request duration"""
    self.duration = time.time() - self.start_time

  def dump_internals(self, extend=None):
    """
    Dumps internal attributes for request.

    :param dict extend: extend dump with additional information
    """
    if extend is None:
      extend = {}
    reqdump = {'eutil':self.eutil, 'db':self.db, 'id':self.id, 'query_id':self.query_id,
               'tool':self.tool, 'url':self.url, 'email':self.contact, 'size':self.size,
               'request_error':self.request_error, 'apikey':'no'}
    if self.apikey is not None:
      reqdump['apikey'] = 'yes'
    return dict(extend, **reqdump)



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: elink_result
   :synopsis: Exports ElinkResult class implementing E-Utils results.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import sys
import json


import entrezpy.base.result
import entrezpy.log.logger


class ElinkResult(entrezpy.base.result.EutilsResult):
  """ The ElinkResult class implements the uniform handling of different Elink
  LinkSets instances. It creates follow-up parameters if possible. ElinkResult
  instances store all results from one Elinker query as an aggregation of
  :class:`entrezpy.elink.linkset.bare.LinkSet` instances.
  The size unit for ElinkResult is :class:`entrezpy.elink.linkset.bare.LinkSet`.

  :param str qid: query id
  :param str cmd: used Elink command
  """

  def __init__(self, qid, cmd):
    """
    :ivar list linksets: list to store analyzed linskets
    :ivar str cmd: invoked ELink command
    """
    super().__init__('elink', qid, db=None)
    self.linksets = []
    self.cmd = cmd
    self.logger = entrezpy.log.logger.get_class_logger(ElinkResult)
    self.logger.debug(json.dumps({'init':self.dump()}))

  def size(self):
    """Implements :meth:`entrezpy.base.result.EutilsResult.size`.
    :rtype: int
    """
    return len(self.linksets)

  def isEmpty(self):
    """Test for empty result

      :returns: True if empty, False otherwise
      :rtype: bool
    """
    if not self.linksets:
      return True
    return False

  def add_linkset(self, linkset):
    """Store linkset in :attr:`self.linkset`

    :param linkset: populated LinkSet
    :type linkset: LinkSet instance
    """
    self.linksets.append(linkset)

  def dump(self):
    """
    :returns: all ELinkResult instance attributes
    :rtype: dict
    """
    return {'function' : self.function, 'size' : self.size(),
            'linksets' : [x.dump() for x in self.linksets]}

  def get_link_parameter(self, reqnum=0):
    """Assemble follow-up parameters depending if the History server has been
    used.

    :returns: parameters for follow-up query
    :rype: dict
    """
    if self.cmd == 'neighbor_history':
      return self.collapse_history_linksets()
    return self.collapse_uid_linksets()

  def collapse_history_linksets(self):
    """Assemble follow-up WebEnv and query_key parameters in linksets.
    Skip those who cannot and test for unexpected result

    :returns: parameters for follow-up query using History server
    :rype: dict
    """
    dbs = {}
    query_keys = []
    for i in self.linksets:
      if not self.canLink(i):
        continue
      for j in i.linkunits:
        dbs[j.db] = 0
        query_keys.append(j.querykey)
    if not dbs:
      return None
    self.check_unexpected_dbnum(dbs)
    if len(query_keys) > 1:
      term = ' OR '.join(str("#{0}".format(x)) for x in query_keys)
      self.logger.debug(json.dumps({'history':{'follow-up term':term}}))
      return {'WebEnv' : self.linksets[0].linkunits[0].webenv,
              'db' : self.linksets[0].linkunits[0].db,
              'term' : term}
    self.logger.debug(json.dumps({'history':{'follow-up querykey':query_keys[0]}}))
    return {'WebEnv' : self.linksets[0].linkunits[0].webenv,
            'db' : self.linksets[0].linkunits[0].db,
            'query_key' : query_keys[0]}

  def collapse_uid_linksets(self):
    """Assemble follow-up UID and database parameters in linksets.
    Skip those who cannot and test for unexpected result

    :returns: parameters for follow-up query using UIDs
    :rype: dict
    """
    dbs = {}
    for i in self.linksets:
      if not self.canLink(i):
        continue
      for j, k in i.get_link_uids().items():
        if j not in dbs:
          dbs[j] = []
        dbs[j] += k
    if not dbs:
      return None
    self.check_unexpected_dbnum(dbs)
    for i in dbs:
      return {'db':i, 'id':dbs[i]}

  def check_unexpected_dbnum(self, dbs):
    """Deal with more databases than expected when linking. Expecting one
    database per request for linking. Abort if more are present since this is
    unexpected.  It shouldn't happen, but make sure to catch such a case, report
    it and abort.

    :param dbs dbs: unique database names encountered in all LinkSets
    """
    if len(dbs) > 1:
      sys.exit(self.logger.error(json.dumps(
       {'Unexpected':'more than 1 dbto in history linking parameter',
        'dbs':[x for x in dbs], 'cmd':self.cmd, 'action':'abort'})))

  def canLink(self, lset):
    """
      Test if linkset can be use to generate automated  follow-up queries

      :param lset: LinkSet
      :type lset: LinkSet instance
      :returns: True if empty, False otherwise
      :rtype: bool
    """
    if lset.canLink:
      return True
    self.logger.warning(json.dumps({f'linkset {lset.category}':'no direct follow-up parameters'}))
    return False



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: eposter
   :synopsis: Exports the Eposter class implementing Epost queries.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import json
import logging

import entrezpy.base.query
import entrezpy.epost.epost_analyzer
import entrezpy.epost.epost_parameter
import entrezpy.epost.epost_request
import entrezpy.log.logger

class Eposter(entrezpy.base.query.EutilsQuery):
  """Eposter implements Epost queries to E-Utilities [0]. EPost posts UIDs to
  the history server. Without passed WebEnv, a new WebEnv and correspndong QueryKey are returned.
  With a given WebEvn the posted UIDs will be added to this WebEnv and the
  corresponding QueryKey is returned. All parameters described in [0] are
  acccepted.
  [0]: https://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.EPost
  """

  def __init__(self, tool, email, apikey=None, apikey_var=None, threads=None, qid=None):
    """Inits Eposter instance with given attributes.

      :param str tool: tool name
      :param str email: user email
      :param str apikey: NCBI apikey
      :param str apikey_var: enviroment variable storing NCBI apikey
      :param int threads: set threads for multithreading
      :param str qid: unique query id
    """
    super().__init__('epost.fcgi', tool, email, apikey, apikey_var, threads, qid)
    self.logger = entrezpy.log.logger.get_class_logger(Eposter)
    self.logger.debug({'init':self.dump()})

  def inquire(self, parameter, analyzer=entrezpy.epost.epost_analyzer.EpostAnalyzer()):
    """Implements :meth:`entrezpy.base.query.inquire` and posts UIDs to Entrez.
       Epost is only one request.

    :param dict parameter: Epost parameter
    :param analyzer analyzer: analyzer for Epost results, default is
        :class:`entrezpy.epost.epost_analyzer.EpostAnalyzer`
    :return: analyzer or None if request errors have been encountered
    :rtype: :class:`entrezpy.base.analyzer.EntrezpyAnalyzer` instance or None
    """
    p = entrezpy.epost.epost_parameter.EpostParameter(parameter)
    self.monitor_start(p)
    self.logger.debug(json.dumps({'parameter':p.dump()}))
    self.add_request(entrezpy.epost.epost_request.EpostRequest(self.eutil, p), analyzer)
    self.request_pool.drain()
    self.monitor_stop()
    if self.isGoodQuery():
      return analyzer
    return None



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.elink.linkset.linked
   :synopsis: Exports LinkedLinkset class implementing 1-to-many Elink results.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""

import entrezpy.elink.linkset.base

class LinkedLinkset(entrezpy.elink.linkset.base.LinkSet):
  """
  The LinkedLinkset class represents a collection of Elink results where
  one UID from a source database (dbfrom) is linked to one or several several
  UID's from a target database (fromdb). This occurs when an elink
  parameter contains several id parameters i.e. id=19880848&id=19822630.

  :param list uidsfrom: UIDs from database to link from
  :param str dbfrom: name of database to link from
  :param bool canLink: linkunits can be used for automated follow-up parameter
  """

  def __init__(self, uidfrom, dbfrom, canLink=True):
    super().__init__('linked', dbfrom, canLink)
    self.uid = int(uidfrom)

  def get_link_uids(self):
    """Collect UIDs in LinksetUnits

    :rtype: dict
    """
    link = {}
    for i in self.linkunits:
      if i.db not in link:
        link[i.db] = []
      link[i.db].append(i.uid)
    return link

  def dump(self):
    """:rtype: dict"""
    return dict({'uid': self.uid}, **self.base_dump())



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: epost_request
  :synopsis: Exports class EpostRequest class implementing individual requests
    from :class:`entrezpy.elink.elink_query.ElinkQuery`

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import json

import entrezpy.base.request
import entrezpy.log.logger


class EpostRequest(entrezpy.base.request.EutilsRequest):
  """ EpostRequest implements a single request as part of an Epost
  query. It stores and prepares the parameters for a single request.
  See :class:`entrezpy.epost.epost_parameter.EpostParameter` for parameter
  description.

  :param parameter: request parameter
  :param type: :class:`entrezpy.epost.epost_parameter.EpostParameter`
  """

  def __init__(self, eutil, parameter):
    super().__init__(eutil, parameter.db)
    self.uids = parameter.uids
    self.size = len(parameter.uids)
    self.webenv = parameter.webenv
    self.retmode = parameter.retmode
    self.logger = entrezpy.log.logger.get_class_logger(EpostRequest)
    self.logger.debug(json.dumps({'init': self.dump()}))

  def get_post_parameter(self):
    """Implements :meth:`entrezpy.base.request.EutilsRequest.get_post_parameter`"""
    return self.prepare_base_qry(extend={'id':','.join(str(x) for x in self.uids),
                                         'WebEnv':self.webenv})

  def dump(self):
    """
    Dump instance attributes

    :rtype: dict
    """
    return self.dump_internals({'retmode':self.retmode, 'WebEnv':self.webenv})



"""
..
  Copyright 2018 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.efetch.efetch_parameter
  :synopsis: Export EfetchParameter implementing Efetch parameters for NCBI
    EUtils Efetch queries

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""

import sys
import math
import json


import entrezpy.base.parameter
import entrezpy.log.logger


# nuccore complexity=> ASN.1 as text
DEF_RETMODE = 'xml'
"""Default retmode for fetch requests"""

class EfetchParameter(entrezpy.base.parameter.EutilsParameter):
  """EfetchParameter implements checks and configures an EftechQuery. A fetch
  query knows its size due to the id parameter or earlier result stored on the
  Entrez history server using WebEnv and query_key. The default retmode
  (fetch format) is set to XML because all E-Utilities can retun XML but not
  JSON, unfortunately.
  """

  req_limits = {'xml' : 10000, 'json' : 500, 'text' : 10000}
  """Max number of UIDs to fetch per request mode"""

  valid_retmodes = {'pmc' : {'xml'},
                    'gene' : {'text', 'xml'},
                    'poset' : {'text', 'xml'},
                    'pubmed' : {'text', 'xml'},
                    'nuccore' : {'text', 'xml'},
                    'protein' : {'text', 'xml'},
                    'sequences' : {'text', 'xml'}}
  """Enforced request uid sizes by NCBI for fetch requests by format"""


  def __init__(self, param):
    super().__init__(param)
    self.uids = param.get('id', [])
    self.rettype = param.get('rettype')
    self.retmode = self.check_retmode(param.get('retmode', DEF_RETMODE))
    self.retmax = self.adjust_retmax(param.get('retmax'))
    self.reqsize = self.adjust_reqsize(param.get('reqsize'))
    self.retstart = int(param.get('retstart', 0))
    self.strand = param.get('strand')
    self.seqstart = param.get('seq_start')
    self.seqstop = param.get('seq_stop')
    self.complexity = param.get('complexity')
    self.calculate_expected_requests(reqsize=self.reqsize)
    self.logger = entrezpy.log.logger.get_class_logger(EfetchParameter)
    self.logger.debug(json.dumps({'init':self.dump()}))

  def adjust_retmax(self, retmax):
    """Adjusts retmax parameter. Order of check is crucial.

    :param int retmax: retmax value
    :return: adjusted retmax or None if all UIDs are fetched
    :rtype: int or None
    """
    if self.uids:           # we got UIDs to fetch
      return len(self.uids)
    if retmax is None:      # we have no clue what to expect, e.g. WebEnv
      self.logger.debug(json.dumps({'No retmax': 'fetching 1 request using retmode/retmax'}))
      return EfetchParameter.req_limits(self.retmode)
    return int(retmax)      # we set a limitation

  def check_retmode(self, retmode):
    """Checks for valid retmode and retmode combination

    :param str retmode: retmode parameter
    :return: retmode
    :rtype: str
    """
    if retmode not in EfetchParameter.req_limits:
      sys.exit(self.logger.error(json.dumps(
        {'Unknown retmode':retmode, 'action':'abort'})))

    if self.db in EfetchParameter.valid_retmodes and \
       retmode not in EfetchParameter.valid_retmodes[self.db]:
      sys.exit(self.logger.error(
        json.dumps({'Bad database retmode':{'db':self.db, 'retmode':retmode},
                    'action':'abort'})))
    return retmode

  def adjust_reqsize(self, reqsize):
    """Adjusts request size for query

    :param reqsize: Request size parameter
    :type  reqsize: str or None
    :return: adjusted request size
    :rtype: int
    """
    if reqsize is None:
      return EfetchParameter.req_limits.get(self.retmode)
    if int(reqsize) > EfetchParameter.req_limits.get(self.retmode):
      return EfetchParameter.req_limits.get(self.retmode)
    if self.retmax and (self.retmax < int(reqsize)):
      return self.retmax
    return int(reqsize)

  def calculate_expected_requests(self, qsize=None, reqsize=None):
    """Calculate anf set the expected number of requests. Uses internal
    parameters if non are provided.

    :param int or None qsize: query size, i.e. expected number of data sets
    :param int reqsize: number of data sets  to fetch in one request
    """
    if not qsize:
      qsize = self.retmax
    if not reqsize:
      reqsize = EfetchParameter.req_limits.get(self.retmode)
    if self.retmax == 0 or (qsize is None):
      qsize = 1
    self.expected_requests = math.ceil(qsize / reqsize)

  def check(self):
    """Implements :class:`entrezpy.base.parameter.EutilsParameter.check` to
    check for the minumum required parameters. Aborts if any check fails.
    """
    if not self.haveDb():
      sys.exit(self.logger.error(json.dumps(
        {'Missing parameter':{'db':self.db}, 'action':'abort'})))

    if not self.haveExpectedRequets():
      sys.exit(self.logger.error(json.dumps(
        {'Bad expected requests':self.expected_requests, 'action':'abort'})))

    if not self.uids and not self.haveQuerykey() and not self.haveWebenv():
      sys.exit(self.logger.error(json.dumps(
        {'Missing parameters':{'id': self.uids, 'QueryKey':self.querykey,
                               'WebEnv':self.webenv},
         'action':'abort'})))

  def dump(self):
    return {'db' : self.db,
            'WebEnv':self.webenv,
            'query_key' : self.querykey,
            'uids' : self.uids,
            'retmode' : self.retmode,
            'rettype' : self.rettype,
            'retstart' : self.retstart,
            'retmax' : self.retmax,
            'strand' : self.strand,
            'seqstart' : self.seqstart,
            'seqstop' : self.seqstop,
            'complexity' : self.complexity,
            'request_size' : self.reqsize,
            'expected_requets' : self.expected_requests}



"""
..
  Copyright 2018 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.esearch.esearch_analyzer
  :synopsis: Exports class EsearchAnalzyer implementing the entrezpy analysis of
    Eutils results from NCBI Eutils queries


.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import json

import entrezpy.base.analyzer
import entrezpy.esearch.esearch_result
import entrezpy.log.logger


class EsearchAnalyzer(entrezpy.base.analyzer.EutilsAnalyzer):
  """EsearchAnalyzer implements the analysis of ESearch responses from E-Utils.
  JSON formatted data is enforced in responses. The result are stored as a
  :class:`entrezpy.esearch.esearch_result.EsearchResult` instance.
  """

  def __init__(self):
    """:ivar result: :class:`entrezpy.esearch.esearch_result.EsearchResult`"""
    super().__init__()
    self.result = None
    self.logger = entrezpy.log.logger.get_class_logger(EsearchAnalyzer)

  def init_result(self, response, request):
    """
    Inits :class:`entrezpy.esearch.esearch_result.EsearchResult`.

    :return: if result is initiated
    :rtype: bool
    """
    if not self.result:
      self.result = entrezpy.esearch.esearch_result.EsearchResult(response, request)
      return True
    return False

  def analyze_result(self, response, request):
    """
    Implements :meth:`entrezpy.base.analyzer.EsearchAnalyzer.analyze_result`.

    :param dict response: Esearch response
    :param request: Esearch request
    :type request: :class:`entrezpy.esearch.esearch_request.EsearchRequest`
    """
    if not self.init_result(response['esearchresult'], request):
      self.result.add_response(response.pop('esearchresult'))

  def analyze_error(self, response, request):
    """
    Implements :meth:`entrezpy.base.analyzer.EutilsAnalyzer.analyze_error`.

      :param dict response: Esearch response
      :param request: Esearch request
      :type request: :class:`entrezpy.esearch.esearch_request.EsearchRequest`
    """
    self.logger.error(json.dumps({'tool':request.tool, 'request':request.id,
      'query':request.query_id, 'error':response['esearchresult'],
      'request-dump':request.dump_internals()}))

  def size(self):
    """
    Returns number of analyzed UIDs in :attr:`.result`

    :rtype: int
    """
    return self.result.size()

  def query_size(self):
    """
    Returns number of expected UIDs in :attr:`.result`

    :rtype: int
    """
    return self.result.query_size()

  def reference(self):
    """
    Returns History Server references from :attr:`.result`

    :return: History Server referencess
    :rtype: :class:`entrezpy.base.referencer.EutilReferencer.Reference`
    """
    return self.result.references.get_reference(self.result.webenv)

  def adjust_followup(self, parameter):
    """
    Adjusts result attributes from follow-up.

    :param parameter: Esearch parameter
    :param type: :class:`entrezpy.esearch.esearch_parameter.EsearchParameter`
    """
    self.result.retmax = parameter.retmax



"""
..
  Copyright 2020 The University of Sydney

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import sys
import atexit
import json
import signal
import queue
import time
import threading

import entrezpy.requester.threadedrequest
import entrezpy.log.logger


class RequestPool:
  """ Threading Pool for requests. This class inits the threading pool,
  adds requests waits until all threads finish. A request consist of a tuple
  with the request and corresponding analyzer. Failed requests are stored
  separately to handle them later. If the number of threads is 0, use
  :meth:`entrezpy.base.query.EutilsQuery.RequestPool.run_single`. Otherwise,
  call :class:`entrezpy.base.query.EutilsQuery.ThreadedRequester`
  This is useful in cases where analyzers are calling not thread-safe methods
  or classes, e.g. Sqlite3
  """

  def __init__(self, num_threads, failed_requests, monitor, requester):
    """
    Initiates a threading pool with a given number of threads.

    :param int num_threads: number of threads
    :param reference failed_requests:
      :attr:`entrezpy.base.query.EutilsQuery.failed_requests`
    :ivar requests: request queue
    :type requests: :class:`queue.Queue`
    """
    self.requests = queue.Queue(num_threads)
    self.failed_requests = failed_requests
    self.requester = requester
    self.monitor = monitor
    self.threads = num_threads
    self.stop_event = threading.Event()
    self.logger = entrezpy.log.logger.get_class_logger(RequestPool)
    self.lock = threading.Lock()
    atexit.register(self.destructor)
    if self.useThreads():
      self.dispatch_workers()
    signal.signal(signal.SIGINT, self.sigint_handler)

  def useThreads(self):
    if self.threads > 0:
      return True
    return False

  def dispatch_workers(self):
    for _ in range(self.threads):
      w = entrezpy.requester.threadedrequest.ThreadedRequester(
        self.requests, self.failed_requests, self.monitor, self.requester, self.stop_event, self.lock)
      w.start()
      self.logger.debug(json.dumps({'thread':w.name, 'status':'started'}))
    self.logger.debug(json.dumps({'threading workers':'dispatched'}))

  def add_request(self, request, analyzer):
    """Adds one request into the threading pool as
    **tuple**\ (`request`, `analzyer`).

    :param  request: entrezpy request instance
    :type   request: :class:`entrezpy.base.request.EutilsRequest`
    :param analyzer: entrezpy analyzer instance
    :type  analyzer: :class:`entrezpy.base.analyzer.EutilsAnalyzer`
    """
    self.requests.put((request, analyzer))

  def drain(self):
    """Empty threading pool and wait until all requests finish"""
    if self.useThreads():
      self.logger.debug(json.dumps({'threads':self.threads}))
      self.logger.debug(json.dumps({'threads':'draining request pool'}))
      self.requests.join()
    else:
      self.logger.debug(json.dumps({'threads':'none'}))
      self.run_single()

  #def check_threads(self, stop_event):
    #while True:
      #for i in threading.enumerate():
        #self.logger.debug(json.dumps({'threads':i}))
        #time.sleep(0.5)

  def run_single(self):
    """Run single threaded requests."""
    while not self.requests.empty():
      request, analyzer = self.requests.get()
      response = self.run_one_request(request)
      if response:
        analyzer.parse(response, request)
      else:
        self.failed_requests.append(request)

  def destructor(self):
    """ Shutdown all ongoing threads when exiting due to an error.

    .. note::
      Deamon processes don't always stop when the main
      program exits and hang aroud. atexit.register(self.desctructor) seems
      to be a way to implement a dectructor. Currently not used.
    """
    pass

  def sigint_handler(self, sigint, frame):
    self.logger.debug(json.dumps({'sigint detected':'stopping threads'}))
    self.stop_event.set()
    self.logger.debug(json.dumps({'sigint detected':'waiting for threads to stop'}))
    time.sleep(1)
    sys.exit(self.logger.info(json.dumps({'sigint detected':'aborting'})))


  def run_one_request(self, request):
    """
    Processes one request from the queue and logs its progress.

    :param request: single entrezpy request
    :type  request: :class:`entrezpy.base.request.EutilsRequest`
    """
    request.start_stopwatch()
    o = self.monitor.get_observer(request.query_id)
    o.observe(request)
    response = self.requester.request(request)
    request.calc_duration()
    o.processed_requests += 1
    #self.logger.debug((response, response.status, response.read(), response.read().decode('utf-8')))
    return response



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.elink.epost_parameter
  :synopsis: Exports class EpostParameter for NCBI E-Utils Esearch queries

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import sys
import json
import logging

import entrezpy.base.parameter
import entrezpy.log.logger


class EpostParameter(entrezpy.base.parameter.EutilsParameter):
  """EpostParameter checks query specific parameters and configures a
  :class:`entrezpy.epost.epost_query.EpostQuery` instance. Force XML since
  Epost responds only XML. Epost requests don't have follow-ups.

  :param dict parameter: Eutils Epost parameters
  """

  def __init__(self, parameter):
    """
    :ivar list uids: UIDs to post
    :ivar str retmode: fix retmode to XML
    :ivar int query_size: size of query, here number of UIDs
    :ivar int request_size: size of request, here nuber if UIDs
    :ivar int expected_requests: number of expected requests, here 1
    """
    super().__init__(parameter)
    self.uids = parameter.get('id', [])
    self.retmode = 'xml'
    self.query_size = len(self.uids)
    self.request_size = self.query_size
    self.expected_requests = 1
    self.check()
    self.logger = entrezpy.log.logger.get_class_logger(EpostParameter)
    self.logger.debug(json.dumps({'init':self.dump()}))

  def check(self):
    """Implements :meth:`entrezpy.base.parameter.EutilsParameter.check`
    by checking for missing database parameter and UIDs.
    """
    if not self.haveDb():
      sys.exit(self.logger.error(json.dumps({'Missing db parameter':'abort'})))
    if self.query_size == 0:
      sys.exit(self.logger.error(json.dumps({'Missing uids':self.uids,
                                             'action':'abort'})))

  def dump(self):
    """Dump instance variables

    :rtype: dict
    """
    return {'db':self.db, 'WebEnv':self.webenv, 'query_key':self.querykey,
            'uids':self.uids, 'retmode':self.retmode, 'doseq':self.doseq,
            'query_size':self.query_size, 'request_size':self.request_size,
            'expected_requets':self.expected_requests}



"""
..
  Copyright 2020 The University of Sydney

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import json
import threading

import entrezpy.requester.requester
import entrezpy.log.logger


class ThreadedRequester(threading.Thread):
  """
  ThreadedRequester handles multitthreaded request. It inherits from
  :class:`threading.Thread`. Requests are fetched  from
  :class:`entrezpy.base.query.EutilsQuery.RequestPool` and processed in
  :meth:`.run`.
  """

  def __init__(self, requests, failed_requests, monitor, requester, stop_event, lock):
    """Inits :class:`.ThreadedRequester` to handle multithreaded requests.

    :param reference requests:
      :attr:`entrezpy.base.query.EutilsQuery.RequestPool.requests`
    :type reference failed_request:
      :attr:`entrezpy.base.query.EutilsQuery.failed_requests`
    """
    super().__init__(daemon=True)
    self.requests = requests
    self.failed_requests = failed_requests
    self.monitor = monitor
    self.requester = requester
    self.stop_event = stop_event
    self.lock = lock
    self.logger = entrezpy.log.logger.get_class_logger(ThreadedRequester)

  def run(self):
    """Overwrite :meth:`threading.Thread.run` for multithreaded requests."""
    while not self.requests.empty() or not self.stop_event.is_set():
      request, analyzer = self.requests.get()
      if self.stop_event.is_set():
        self.logger.info(json.dumps({'received stop signal':'aborting'}))
        break
      self.run_one_request(request, analyzer)
      self.logger.info(json.dumps({'query':request.query_id,
                                   'request':request.id,
                                   'status': request.status}))
      self.requests.task_done()

  def run_one_request(self, request, analyzer):
    """
    Processes one request from the queue and logs its progress.

    :param request: single entrezpy request
    :type  request: :class:`entrezpy.base.request.EutilsRequest`
    """
    self.lock.acquire()
    self.logger.debug(json.dumps({'lock':self.lock.locked()}))
    request.start_stopwatch()
    o = self.monitor.get_observer(request.query_id)
    o.observe(request)
    response = self.requester.request(request)
    request.calc_duration()
    o.processed_requests += 1
    self.logger.debug(response)
    if response:
      analyzer.parse(response, request)
    else:
      self.failed_requests.append(request)
    self.lock.release()
    self.logger.debug(json.dumps({'lock':self.lock.locked()}))



"""
..
  Copyright 2018 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.esearch.esearcher
   :synopsis: Exports class Esearcher implementing entrezpy Esearch queries to
      NCBI EUtils Esearch

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""

import sys
import json

import entrezpy.base.query
import entrezpy.esearch.esearch_parameter
import entrezpy.esearch.esearch_analyzer
import entrezpy.esearch.esearch_request
import entrezpy.log.logger


class Esearcher(entrezpy.base.query.EutilsQuery):
  """
  Esearcher implements ESearch queries to NCBI's E-Utilities. Esearch queries
  return UIDs or WebEnv/QueryKey references to Entrez' History server.
  Esearcher implments :meth:`entrezpy.base.query.EutilsQuery.inquire` which
  analyzes the first result and automatically configures subseqeunt requests to
  get all queried UIDs if required.
  """

  def __init__(self, tool, email, apikey=None, apikey_var=None, threads=None, qid=None):
    super().__init__('esearch.fcgi', tool, email, apikey=apikey, threads=threads, qid=qid)
    self.logger = entrezpy.log.logger.get_class_logger(Esearcher)
    self.logger.debug(json.dumps({'init':self.dump()}))

  def inquire(self, parameter, analyzer=entrezpy.esearch.esearch_analyzer.EsearchAnalyzer()):
    """Implements :meth:`entrezpy.base.query.EutilsQuery.inquire` and configures
    follow-up requests if required.

    :param dict parameter: ESearch parameter
    :param analyzer analyzer: analyzer for ESearch results, default is
      :class:`entrezpy.esearch.esearch_analyzer.EsearchAnalyzer`
    :return: analyzer instance or None if request errors have been encountered
    :rtype: :class:`entrezpy.esearch.esearch_analyzer.EsearchAnalyzer` or None
    """
    p = entrezpy.esearch.esearch_parameter.EsearchParameter(parameter)
    self.logger.debug(json.dumps({'parameter':p.dump()}))
    self.monitor_start(p)
    follow_up = self.initial_search(p, analyzer)
    if not follow_up:
      self.monitor_stop()
      if not analyzer.isSuccess():
        return None
      return analyzer
    self.logger.debug(json.dumps({'Follow-up': follow_up.dump()}))
    self.monitor_update(follow_up)
    req_size = follow_up.reqsize
    for i in range(1, follow_up.expected_requests):
      if (i * req_size + req_size) > follow_up.retmax:
        self.logger.debug(json.dumps({'adjust-reqsize':
          {'request':i, 'start':(i*follow_up.reqsize), 'end':i*req_size+req_size,
           'query_size':follow_up.reqsize, 'adjusted-reqsize':follow_up.retmax%req_size}}))
        req_size = follow_up.retmax % req_size
      self.add_request(entrezpy.esearch.esearch_request.EsearchRequest(self.eutil,
                                                                       follow_up,
                                                                       (i*follow_up.reqsize),
                                                                       req_size), analyzer)
      self.logger.debug(json.dumps({'added request':i, 'reqsize':req_size,
        'expected':follow_up.expected_requests,'start':(i*follow_up.reqsize),
        'end':(i*follow_up.reqsize)+req_size}))
    self.request_pool.drain()
    self.monitor_stop()
    if not self.isGoodQuery():
      self.logger.error(json.dumps({'follow-up':failed}))
      return None
    return analyzer

  def initial_search(self, parameter, analyzer):
    """Does first request and triggers follow-up if required or possible.

    :param parameter: Esearch parameter instances
    :type  parameter: :class:`entrezpy.esearch.esearch_parameter.EsearchParamater`
    :param analyzer: Esearch analyzer instance
    :type  analyzer: :class:`entrezpy.esearch.esearch_analyzer.EsearchAnalyzer`
    :return: follow-up parameter or None
    :rtype: :class:`entrezpy.esearch.esearch_parameter.EsearchParamater` or None
    """
    self.add_request(entrezpy.esearch.esearch_request.EsearchRequest(self.eutil,
                                                                     parameter,
                                                                     parameter.retstart,
                                                                     parameter.reqsize), analyzer)
    self.request_pool.drain()
    if not self.isGoodQuery():
      self.logger.error(json.dumps({'initial request':'failed'}))
      return None
    if not analyzer.isSuccess():
      self.logger.error(json.dumps({'initial response':'failed'}))
      return None
    if not parameter.uilist: # we care only about count
      self.logger.debug(json.dumps({'initial response':'no uilist'}))
      return None
    if parameter.retmax == 0: # synonym for uilist
      self.logger.debug(json.dumps({'initial response':'retmax == 0'}))
      return None
    if reachedLimit(parameter, analyzer):# reached limit in first search
      self.logger.debug(json.dumps({'initial response':'fetched whole query'}))
      return None
    self.logger.debug(json.dumps({'initial response': 'follow-up required'}))
    return configure_follow_up(parameter, analyzer) # We need mooaahhr

  def isGoodQuery(self):
    """
    Tests for request errors

      :rtype: bool
    """
    if not self.failed_requests:
      self.logger.info(json.dumps({'query': self.id, 'status' : 'OK'}))
      return True
    self.logger.warning(json.dumps({'query': self.id, 'status' : 'failed'}))
    self.logger.debug(json.dumps({'query': self.id, 'status' : 'failed',
      'request-dumps' : [x.dump_internals() for x in self.failed_requests]}))
    return False

def configure_follow_up(parameter, analyzer):
  """
  Adjusting EsearchParameter to follow-up results based on the initial Esearch
  result. Fetch remaining UIDs using the history server.

  :param analyzer: Esearch analyzer instance
  :type  analyzer: :class:`entrezpy.search.esearch_analyzer.EsearchAnalyzer`
  :param parameter: Initial Esearch parameter
  :type  result: :class:`entrezpy.search.esearch_parameter.EsearchParameter`
  """

  parameter.term = None
  if not parameter.retmax:
    parameter.retmax = analyzer.query_size()
  analyzer.adjust_followup(parameter)
  parameter.webenv = analyzer.reference().webenv
  parameter.querykey = analyzer.reference().querykeys[0]
  parameter.calculate_expected_requests(reqsize=parameter.reqsize)
  parameter.check()
  return parameter

def reachedLimit(parameter, analyzer):
  """
  Checks if the set limit has been reached

  :rtype: bool
  """
  if analyzer.query_size() == analyzer.size(): # fetched all UIDs
    return True
  if not parameter.retmax:  # We have no limit
    return False
  if analyzer.size() == parameter.retmax: # Fetched limit set by retmax
    return True
  return False



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.elink.linkset.unit.linkin
  :synopsis: Exports class LinkIn implementing Elink results from the Elink
    `ncheck` command reporting if links are available within the same Entrez
    database.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import entrezpy.elink.linkset.unit.linksetunit


class LinkIn(entrezpy.elink.linkset.unit.linksetunit.LinksetUnit):
  """ The `LinkIn` class represents a result from the `ncheck` Elink command.
  Results show if links are available within the same Entrez databases.

  :param str dbto: name of target database
  :param str hasneighbor: Y or N indicating if links are avaible
  """

  @classmethod
  def new(cls, dbto, hasneighbor):
    """:rtype: :class:`linkin.LinkIn`"""
    return cls(dbto, hasneighbor)

  def __init__(self, dbto=None, hasneighbor=None):
    """:ivar bool hasneighbor:"""
    super().__init__(dbto, None)
    self.hasneighbor = bool(hasneighbor == 'Y')

  def dump(self):
    """:rtype: dict"""
    return dict({'hasneighbor' : self.hasneighbor}, **self.basic_dump())



"""
..
  Copyright 2018 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.esearch.esearch_request
  :synopsis: Exports class SearchRequest implementing individual entrezpy
    requests to NCBI Eutils Esearch


.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import json

import entrezpy.base.request
import entrezpy.log.logger


class EsearchRequest(entrezpy.base.request.EutilsRequest):
  """
  The EsearchRequest class implements a single request as part of a Esearch
  query. It stores and prepares the parameters for a single request.
  See :class:`entrezpy.elink.elink_parameter.ElinkParameter` for parameter
  description. Requests sizes are congifured from setting a start, i.e. the
  index of the first UID to fetch, and its size, i.e. how many to fetch.
  These are set by :meth:`entrezpy.esearch.esearch_query.Esearcher.inquire`.

  :param parameter: request parameter
  :param type: :class:`entrezpy.elink.elink_parameter.ElinkParameter`
  :param int start: number of first UID to fetch
  :param int size: requets size
  """

  def __init__(self, eutil,  parameter, start, size):
    super().__init__(eutil, parameter.db)
    self.id = None
    self.retstart = start
    self.retmax = size
    self.term = parameter.term
    self.retmode = parameter.retmode
    self.usehistory = parameter.usehistory
    self.webenv = parameter.webenv
    self.querykey = parameter.querykey
    self.sort = parameter.sort
    self.field = parameter.field
    self.idtype = parameter.idtype
    self.datetype = parameter.datetype
    self.reldate = parameter.reldate
    self.mindate = parameter.mindate
    self.maxdate = parameter.maxdate
    self.logger = entrezpy.log.logger.get_class_logger(EsearchRequest)
    self.logger.debug(json.dumps({'init': self.dump()}))

  def get_post_parameter(self):
    qry = self.prepare_base_qry()
    if self.term:
      qry.update({'term' : self.term})
    if self.retmax or self.retmax == 0:
      qry.update({'retmax' : self.retmax})
    if self.retstart:
      qry.update({'retstart' : self.retstart})
    if self.retmode:
      qry.update({'retmode' : self.retmode})
    if self.usehistory:
      qry.update({'usehistory' : 'y'})
    if self.webenv:
      qry.update({'WebEnv' : self.webenv})
    if self.querykey:
      qry.update({'query_key' : self.querykey})
    if self.sort:
      qry.update({'sort' : self.sort})
    if self.field:
      qry.update({'field' : self.field})
    if self.datetype:
      qry.update({'datetype' : self.datetype})
    if self.reldate:
      qry.update({'reldate' : self.reldate})
    if self.mindate:
      qry.update({'mindate' : self.mindate})
    if self.maxdate:
      qry.update({'maxdate' : self.maxdate})
    if self.idtype:
      qry.update({'idtype' : self.idtype})
    self.logger.debug(json.dumps({'POST': self.dump()}))
    return qry

  def dump(self):
    """:rtype: dict"""
    return self.dump_internals({'retmax':self.retmax, 'WebEnv':self.webenv,
      'retmode':self.retmode, 'usehistory':self.usehistory, 'sort':self.sort,
      'retstart':self.retstart, 'query_key' : self.querykey, 'field':self.field,
      'datetype':self.datetype, 'reldate':self.reldate, 'mindate':self.mindate,
      'maxdate':self.maxdate, 'id':self.id, 'query-id':self.query_id})



"""
..
  Copyright 2018 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.esummary.esummary_result
  :synopsis: Exports class EsummaryResult implementing entrezpy results from
    NCBI Esummary E-Utility requests

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import json

import entrezpy.base.result
import entrezpy.log.logger

class EsummaryResult(entrezpy.base.result.EutilsResult):
  """
  EsummaryResult stores summaries in :attr:`.summaries`, avoiding duplicates and
  quick access. EsummaryResult has no WebEnv references.
  """

  def __init__(self, response, request):
    super().__init__('esummary', request.query_id, request.db)
    self.summaries = {}
    if response:
      self.add_summaries(response['result'])
    self.logger = entrezpy.log.logger.get_class_logger(EsummaryResult)
    self.logger.debug(json.dumps({'init':self.dump()}))

  def dump(self):
    """:rtype: dict"""
    return {'db':self.db, 'size':self.size(), 'function':self.function,
            'summaries':[self.summaries[x] for x in self.summaries]}

  def get_link_parameter(self, reqnum=0):
    """
    Esummary has no link automated link ability.

    :return: None
    """
    return None

  def size(self):
    return len(self.summaries)

  def isEmpty(self):
    if self.size() == 0:
      return True
    return False

  def add_summaries(self, results):
    """
    Adds summaries form a Esummary E-Utiliy response.

    :param dict results: Esummaries
    """
    if results:
      for i in results['uids']:
        if int(i) not in self.summaries:
          self.summaries[int(i)] = results.get(i)



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.epost.epost_analyzer
  :synopsis: Exports class EpostAnalzyer implementing the analysis of Epost
    query results.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import json
import xml.etree.ElementTree

import entrezpy.base.analyzer
import entrezpy.epost.epost_result
import entrezpy.log.logger


class EpostAnalyzer(entrezpy.base.analyzer.EutilsAnalyzer):
  """EpostAnalyzer implements the analysis of EPost responses from E-Utils.
  Epost puts UIDs onto the History server and returnes the corresponding
  WebEnv and QueryKey. Epost does only XML as response, therefore a dictionary
  imitating a JSON input is assembled and passed as result to
  :class:`entrezpy.epost.epost_result.EpostResult`
  """

  def __init__(self):
    super().__init__()
    self.logger = entrezpy.log.logger.get_class_logger(EpostAnalyzer)

  def init_result(self, response, request):
    """Implements :meth:`entrezpy.base.analyzer.EutilsAnalyzer.init_result` and
       inits :class:`entrezpy.epost.epost_result.EpostResult`.
    """
    if not self.result:
      self.result = entrezpy.epost.epost_result.EpostResult(response, request)

  def analyze_result(self, response, request):
    """Implements :meth:`entrezpy.base.analyzer.EutilsAnalyzer.analyze_result`.
      The response is one WebEnv and QueryKey and the result can be initiated
      after parsing them.

      :param response: EUtils response
      :param request: entrezpy request
    """
    epost_res = {}
    for event, elem in xml.etree.ElementTree.iterparse(response, events=["end"]):
      if event == 'end' and elem.tag == 'QueryKey':
        epost_res['querykey'] = int(elem.text)
      if event == 'end' and elem.tag == 'WebEnv':
        epost_res['webenv'] = elem.text
      elem.clear()
    self.init_result(epost_res, request)

  def analyze_error(self, response, request):
    """Implements :meth:`entrezpy.base.analyzer.EutilsAnalyzer.analyze_error`.

      :param response: EUtils response
      :param request: entrezpy request
    """
    error = None
    for _, elem in xml.etree.ElementTree.iterparse(response, events=["end"]):
      if elem.tag == 'ERROR':
        error = elem.text
        break
      elem.clear()
    self.logger.error(json.dumps({'tool':request.tool, 'error':error,
                                  'requestid':request.id, 'queryid':request.query_id,
                                  'request-dump':request.dump_internals()}))



"""
..
  Copyright 2018, 2019 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.elinker.elink_analyzer
   :synopsis: Exports the ElinkeAnalzyer class to analyze Elink query results.

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import sys
import json
import xml.etree.ElementTree


import entrezpy.base.analyzer
import entrezpy.elink.elink_result
import entrezpy.elink.linkset.base
import entrezpy.elink.linkset.linked
import entrezpy.elink.linkset.relaxed
import entrezpy.elink.linkset.unit.linkin
import entrezpy.elink.linkset.unit.linklist
import entrezpy.elink.linkset.unit.linkout
import entrezpy.elink.linkset.unit.linkoutallattribute
import entrezpy.elink.linkset.unit.linkoutnonlibattribute
import entrezpy.elink.linkset.unit.linkoutprovider
import entrezpy.elink.linkset.unit.neighbor
import entrezpy.elink.linkset.unit.neighborhistory
import entrezpy.elink.linkset.unit.neighborscore
import entrezpy.log.logger


class ElinkAnalyzer(entrezpy.base.analyzer.EutilsAnalyzer):
  """
  ElinkAnalyzer implements parsing and superficial analysis of responses from
  ELink queries. ElinkAnalyzer implements the virtual methods
  :meth:`.analyze_result` and :meth:`.analyze_error`. The variety in possible
  Elink response formats results in several specialized parser. Default is to
  obtain results in JSON.

  ElinkAnalyzer instances create :class:`linked.Linkset` or
  :class:`relaxed.Linkset` instances, depending on the request Elink result.
  :func:`entrezpy.elink.linkset.bare.Linkset.new_unit` is called to set the
  type of LinkSet unit based ont he used Elink command.

  .. warning:: Expect for 'llinkslib', all responses are expected in JSON.
                 ElinkAnalyzer will abort if a response from another command is
                 not in JSON.
  """

  def __init__(self):
    """:ivar result: :class:`entrezpy.elink.elink_result.ElinkResult`"""
    super().__init__()
    self.logger = entrezpy.log.logger.get_class_logger(ElinkAnalyzer)
    self.isLinkList = False

  def init_result(self, response, request):
    """Inits :class:`entrezpy.elink.elink_result.ElinkResult`"""
    if not self.result:
      self.result = entrezpy.elink.elink_result.ElinkResult(request.query_id, request.cmd)

  def analyze_error(self, response, request):
    """Implements virtual function :func:`entrezpy.base.analyzer.analyze_error`."""
    self.logger.debug(json.dumps({'response':response, 'request-dump':request.dump_internals()}))

  def get_linkset_unit(self, elink_cmd):
    if elink_cmd == 'neighbor':
      return entrezpy.elink.linkset.unit.neighbor.Neighbor()
    if elink_cmd == 'neighbor_score':
      return entrezpy.elink.linkset.unit.neighborscore.NeighborScore()
    if elink_cmd == 'neighbor_history':
      return entrezpy.elink.linkset.unit.neighborhistory.NeighborHistory()

    self.isLinkList = True
    if elink_cmd == 'acheck':
      return entrezpy.elink.linkset.unit.linklist.LinkList()
    if elink_cmd == 'ncheck':
      return entrezpy.elink.linkset.unit.linkin.LinkIn()
    if elink_cmd == 'lcheck':
      return entrezpy.elink.linkset.unit.linkout.LinkOut()
    if elink_cmd == 'llinks':
      return entrezpy.elink.linkset.unit.linkoutnonlibattribute.LinkOutNonlibAttributes()
    if elink_cmd == 'llinkslib':
      return entrezpy.elink.linkset.unit.linkoutallattribute.LinkOutAllAttribute()
    if elink_cmd == 'prlinks':
      return entrezpy.elink.linkset.unit.linkoutprovider.LinkOutProvider()

  def analyze_result(self, response, request):
    """
    Implements virtual method :meth:`entrezpy.base.analyzer.analyze_result` and
    checks used elink command to run according result parser.
    """
    self.init_result(response, request)
    if request.cmd != 'llinkslib' and request.retmode != 'json':
      sys.exit(self.logger.error(json.dumps({'unknown cmd':request.cmd,
                                             'format':request.retmode,
                                             'action':abort})))
    if request.retmode == 'json':
      self.logger.debug(json.dumps({'response':response}))
    else:
      self.logger.debug(json.dumps({'response':response.getvalue()}))

    lset_unit = self.get_linkset_unit(request.cmd)
    if request.cmd == 'llinkslib':  # only available as XML, groan
      self.parse_llinkslib(response, lset_unit)
    elif not self.isLinkList:
      self.analyze_links(response['linksets'], lset_unit)
    elif self.isLinkList:
      self.analyze_linklist(response['linksets'], lset_unit)
    else:
      sys.exit(self.logger.error(json.dumps({'unknown elink cmd':request.cmd,
                                             'action':'abort'})))

  def analyze_linklist(self, linksets, lset_unit):
    """
    Parses ELink responses listing link information for UIDs.

    :param dict linksets: 'linkset' part in an ELink JSON response from NCBI.
    :param lset_unit: Elink LinkSet unit instance
    """
    for i in linksets:
      if 'idurllist' in i:
        for j in i.get('idurllist'):
          lset = entrezpy.elink.linkset.linked.LinkedLinkset(j['id'], i['dbfrom'], canLink=False)
          self.logger.debug(json.dumps({'created':lset.dump()}))
          for k in j['objurls']:
            lset.add_linkunit(lset_unit.new(k))
            self.logger.debug(json.dumps({'added':lset.linkunits[-1].dump()}))
          self.result.add_linkset(lset)
      elif 'idlinksets' in i['idchecklist']:
        for j in i['idchecklist'].get('idlinksets'):
          lset = entrezpy.elink.linkset.linked.LinkedLinkset(j['id'], i['dbfrom'], canLink=False)
          self.logger.debug(json.dumps({'created':lset.dump()}))
          for k in j['linkinfos']:
            lset.add_linkunit(lset_unit.new(k['dbto'], k['linkname'], k.get('menutag'),
                                            k.get('htmltag'), k.get('priority')))
            self.logger.debug(json.dumps({'added':lset.linkunits[-1].dump()}))
          self.result.add_linkset(lset)
      elif 'ids' in i['idchecklist']:
        for j in i['idchecklist'].get('ids'):
          lset = entrezpy.elink.linkset.linked.LinkedLinkset(j['value'], i['dbfrom'], canLink=False)
          self.logger.debug(json.dumps({'created':lset.dump()}))
          if 'hasneighbor' in j:
            lset.add_linkunit(lset_unit.new(i['dbfrom'], j['hasneighbor']))
            self.logger.debug(json.dumps({'added':lset.linkunits[-1].dump()}))
          if 'haslinkout' in j:
            lset.add_linkunit(lset_unit.new(i['dbfrom'], j['haslinkout']))
            self.logger.debug(json.dumps({'added':lset.linkunits[-1].dump()}))
          self.result.add_linkset(lset)

  def analyze_links(self, linksets, lset_unit):
    """
    Parses ELink responses with links  as UIDs or History server references.

    :param dict linksets: 'linkset' part in an ELink JSON response from NCBI.
    :param lset_unit: Elink LinkSet unit instance
    """
    for i in linksets:
      #assume 1-to-many link as default
      lset = entrezpy.elink.linkset.linked.LinkedLinkset(i['ids'][0], i['dbfrom'])
      if len(i['ids']) > 1: # OK, it's many-to-many
        lset = entrezpy.elink.linkset.relaxed.RelaxedLinkset(i['ids'], i['dbfrom'])
      self.logger.debug(json.dumps({'created':lset.dump()}))
      if 'linksetdbs' in i:
        for j in i['linksetdbs']:
          if 'ERROR' in j:
            self.logger.error(json.dumps({'response':j['ERROR']}))
            self.hasErrorResponse = True
          if 'links' in j:
            for k in j['links']:
              lset.add_linkunit(lset_unit.new(k, j['dbto'], j['linkname']))
              self.logger.debug(json.dumps({'added':lset.linkunits[-1].dump()}))
      elif 'linksetdbhistories' in i:
        for j in i['linksetdbhistories']:
          lset.add_linkunit(lset_unit.new(j['dbto'], j['linkname'], i['webenv'], j['querykey']))
          self.logger.debug(json.dumps({'added':lset.linkunits[-1].dump()}))
      else:
        self.logger.warning(json.dumps({'Empty linkset'}))
        continue
      self.result.add_linkset(lset)

  def parse_llinkslib(self, response, lset_unit, lset=None):
    """
    Exclusive XML parser for 'llinkslib' responses. Its approach is ugly but
    parses the XML. The cmd parametes is always 'llinkslib' but retains the
    calling signature.

    :param io.StringIO response: XML response from Entrez
    :param lset_unit: Elink LinkSet unit instance
    """
    provider_tagmap = {'iconurl', 'id', 'url', 'name', 'nameabbr'}
    unit_tagmap = {'iconurl', 'url', 'name', 'nameabbr', 'attribute',
                   'category', 'linkname', 'subjecttype'}
    unit = {'attributes' : [], 'provider' : {}}
    dbfrom = None
    isLinkset = False
    isObjurl = False
    isProvider = False
    for event, elem in xml.etree.ElementTree.iterparse(response, events=['start','end']):
      if event == 'start':
        if elem.tag == 'LinkSet':
          isLinkset = True
        if elem.tag == 'ObjUrl':
          isObjurl = True
          unit = {'attributes':[], 'provider':{}}

      if event == 'end':
        if elem.tag == 'LinkSet':
          isLinkset = False
        if elem.tag == 'ObjUrl':
          isObjurl = False
          lset.add_linkunit(lset_unit.new(unit))
          self.logger.debug(json.dumps({'added':lset.linkunits[-1].dump()}))
        if elem.tag == 'IdUrlSet':
          self.result.add_linkset(lset)

      if isLinkset and not isObjurl:
        if event == 'end' and elem.tag == 'DbFrom':
          dbfrom = elem.text
        if elem.tag == 'Id' and not isObjurl:
          lset = entrezpy.elink.linkset.linked.LinkedLinkset(int(elem.text), dbfrom, canLink=False)
          self.logger.debug(json.dumps({'created':lset.dump()}))

      if isLinkset and isObjurl:
        if elem.tag == 'Provider':
          isProvider = bool(event == 'start')

        if event == 'end' and elem.tag.lower() in unit_tagmap:
          if elem.tag.lower() == 'attribute':
            unit['attributes'].append(elem.text)
          elif isProvider and elem.tag.lower() in provider_tagmap:
            unit['provider'][elem.tag.lower()] = elem.text
          else:
            unit[elem.tag.lower()] = elem.text
          elem.clear()



"""
..
  Copyright 2018 The University of Sydney
  This file is part of entrezpy.

  Entrezpy is free software: you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option) any
  later version.

  Entrezpy is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with entrezpy.  If not, see <https://www.gnu.org/licenses/>.

.. module:: entrezpy.base.analyzer
  :synopsis: Exports the base class for entrezpy analzyer for NCBI E-Utils
    responses

.. moduleauthor:: Jan P Buchmann <jan.buchmann@sydney.edu.au>
"""


import io
import sys
import json
import xml.etree.ElementTree

import entrezpy.log.logger


class EutilsAnalyzer:
  """EutilsAnalyzer is the base class for an entrezpy analyzer.
  It prepares the response based on the requested format and checks for
  E-Utilities errors. The function parse() is invoked after every request by the
  corresponding query class, e.g. Esearcher. This allows analyzing data as it
  arrives without waiting until larger queries have been fetched. This approach
  allows implementing analyzers which can store already downloaded data to
  establish checkpoints or trigger other actions based on the received data.

  Two virtual classes are the core and need their own implementation to support
  specific queries:

  - :meth:`.analyze_error`
  - :meth:`.analyze_result`

  .. note::
    Responses from NCBI are not very well documented and functions will
    be extended as new errors are encountered."""

  known_fmts = {'xml', 'json', 'text'}
  """Store formats known to EutilsAnalzyer"""

  def __init__(self):
    """Inits EutilsAnalyzer with unknown type of result yet. The result needs to
    be set upon receiving the first response by :meth:`.init_result`.

    :ivar bool hasErrorResponse: flag indicating error in response
    :ivar result: result instance
    :type result: :class:`entrezpy.base.result.EutilsResult`
    """
    self.hasErrorResponse = False
    self.result = None
    self.logger = entrezpy.log.logger.get_class_logger(EutilsAnalyzer)

  def init_result(self, response, request):
    """Virtual function to initialize result instance. This allows to set
    attributes from the first response and request.

    :param response: converted response from :meth:`.convert_response`
    :type  response: dict or `io.StringIO`
    :raises NotImplementedError: if implementation is missing
    """
    raise NotImplementedError("Require implementation of analyze_error()")

  def analyze_error(self, response, request):
    """Virtual function to handle error responses

    :param response: converted response from :meth:`.convert_response`
    :type  response: dict or `io.StringIO`
    :raises NotImplementedError: if implementation is missing
    """
    raise NotImplementedError("Require implementation of analyze_error()")

  def analyze_result(self, response, request):
    """Virtual function to handle responses, i.e. parsing them and prepare
    them for :class:`entrezpy.base.result.EutilsResult`

    :param response: converted response from :meth:`.convert_response`
    :type  response: dict or `io.StringIO`
    :raises NotImplementedError: if implementation is missing
    """
    raise NotImplementedError("Require implementation of analyze_result()")

  def parse(self, raw_response, request):
    """Check for errors and calls parser for the raw response.

    :param raw_response: response from :class:`entrezpy.requester.requester.Requester`
    :type  raw_response: :class:`urllib.request.Request`
    :param request: query request
    :type  request: :class:`entrezpy.base.request.EutilsRequest`
    :raises NotImplementedError: if request format is not in
      :attr:`EutilsAnalyzer.known_fmts`"""
    if request.retmode not in EutilsAnalyzer.known_fmts:
      self.logger.error(json.dumps({'unknown format': request.retmode}))
      raise NotImplementedError(f"Unknown format: {request.retmode}")
    #raw_response_decoded = raw_response.read().decode('utf-8')
    #self.logger.info((raw_response, raw_response.status, raw_response.read(), raw_response_decoded))
    #self.logger.info(json.dumps({'parsing':raw_response_decoded, 'req':request.dump()}))
    response = self.convert_response(raw_response.read().decode('utf-8'), request)
    if self.isErrorResponse(response, request):
      self.hasErrorResponse = True
      self.analyze_error(response, request)
    else:
      self.analyze_result(response, request)

    if self.result is None:
      sys.exit(json.dumps({'result attribute': 'not set','action' : 'abort'}))

  def convert_response(self, raw_response_decoded, request):
    """
    Converts raw_response into the expected format, deduced from request and
    set via the retmode parameter.

    :param raw_response:  response :class:`entrezpy.requester.requester.Requester`
    :type  raw_response: :class:`urllib.request.Request`
    :param request: query request
    :type  request: :class:`entrezpy.base.request.EutilsRequest`
    :return: response in parseable format
    :rtype: dict or :class:`io.stringIO`

    ..note::
      Using threads without locks randomly 'looses' the response, i.e. the
      raw response is emptied between requests. With locks, it works, but
      threading is not much faster than non-threading. It seems JSON is more
      prone to this than XML.
    """
    if request.retmode == 'json':
      #self.logger.info(json.dumps({'raw':raw_response_decoded, 'req':request.dump()}))
      return json.loads(raw_response_decoded)
    return io.StringIO(raw_response_decoded)

  def isErrorResponse(self, response, request):
    """Checking for error messages in response from Entrez Servers and set flag
    :attr:`.hasErrorResponse`.

    :param response: parseable response from :meth:`.convert_response`
    :type  response: dict or :class:`io.stringIO`
    :param request: query request
    :type  request: :class:`entrezpy.base.request.EutilsRequest`
    :return: error status
    :rtype: bool"""
    if request.retmode == 'xml':
      hasErrorResponse = self.check_error_xml(response)
      response.seek(0)
      return hasErrorResponse
    if request.retmode == 'json':
      hasErrorResponse = self.check_error_json(response)
    return hasErrorResponse

  def check_error_xml(self, response):
    """Checks for errors in XML responses

    :param response: XML response
    :type response: :class:`io.stringIO`
    :return: if XML response has error message
    :rtype: bool"""
    for _, elem in xml.etree.ElementTree.iterparse(response, events=["end"]):
      if elem.tag == 'ERROR':
        elem.clear()
        return True
      elem.clear()
    return False

  def check_error_json(self, response):
    """Checks for errors in JSON responses. Not unified among Eutil functions.

    :param dict response: reponse
    :return: status if JSON response has error message
    :rtype: bool"""
    if response['header']['type'] == 'esearch' and 'ERROR' in response['esearchresult']:
      return True
    if response['header']['type'] == 'elink' and 'ERROR' in response:
      return True
    if 'esummaryresult' in response:
      if response['esummaryresult'][0].split(' ')[0] == 'Invalid':
        return True
    if 'error' in response:
      return True
    return False

  def isSuccess(self):
    """Test if response has errors

    :rtype: bool"""
    if self.hasErrorResponse:
      return False
    return True

  def get_result(self):
    """Return result

    :return: result instance
    :rtype: :class:`entrezpy.base.result.EutilsResult`
    """
    return self.result

  def follow_up(self):
    """Return follow-up parameters if available

    :return: Follow-up parameters
    :rtype: dict
    """
    return self.result.get_link_parameter()

  def isEmpty(self):
    """Test for empty result

    :rtype: bool
    """
    if self.result.size() == 0:
      return True
    return False
